{-# LANGUAGE NamedFieldPuns #-}
module Data.Metadata.Trie
  () where

import Data.Foldable (toList)
import Data.Map (Map)
import qualified Data.Map as Map (delete, empty, insert, lookup, singleton)

data Trie e n =
  Trie
    { store :: Maybe n
    , edges :: Edges e n }
type Edges e n = Map e (Trie e n)

data Zipper e n =
    Top
  | Zip
      { atValue :: Maybe n
      , otherEdges :: Edges e n
      , byEdge :: e
      , above :: Zipper e n }

type EditState e n = (Trie e n, Zipper e n)

empty :: Trie e n
empty = Trie { store = Nothing, edges = Map.empty }

zipUp :: Ord e => EditState e n -> Trie e n
zipUp (t, Top) = t
zipUp (t, Zip {atValue, otherEdges, byEdge, above}) =
  zipUp
    ( Trie { store = atValue, edges = Map.insert byEdge t otherEdges }
    , above )

trieOf :: Foldable t => t e -> n -> Trie e n
trieOf path n = foldr insertAt (Trie { store = Just n, edges = Map.empty }) path
  where
    insertAt e t = Trie { store = Nothing, edges = Map.singleton e t }

at :: (Foldable t, Ord e) => Trie e n -> t e -> Maybe n
at trie = at_ trie . toList
  where
    at_ (Trie {store}) [] = store
    at_ (Trie {edges}) (e:es) = Map.lookup e edges >>= (`at_` es)

insert :: (Ord e, Foldable t) => t e -> n -> Trie e n -> Trie e n
insert path n trie = zipUp (edit (toList path) (trie, Top))
  where
    edit [] (trie, zipper) = (trie {store = Just n}, zipper)
    edit (e:es) (Trie {store, edges}, above) =
      let (subTree, otherEdges) = partition e edges in
      edit es (subTree, Zip {atValue = store, otherEdges, byEdge = e, above})
    partition e edges =
      case Map.lookup e edges of
        Just subTree -> (subTree, Map.delete e edges)
        _ -> (empty, Map.empty)

type Indexed e n = ([e], n)

index :: (Ord e, Foldable t, Functor t) => (a -> Indexed e n) -> t a -> Trie e n
index by = foldr (uncurry insert) empty . fmap by
